"""
Copyright or © or Copr. University of Nantes / LS2N (2021)

Contributor:
* Alexis Bitaillou

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
"""

import copy
import re
import statistics
from typing import List, Union, Callable

def convert_time_unit(origin_interval: float, origin_unit: str) -> float:
    if origin_unit.upper() == "MS":
        multiplicator = 1 / 1000
    elif origin_unit.upper() == "US":
        multiplicator = 1 / 1000000
    elif origin_unit.upper() == "S":
        multiplicator = 1
    elif origin_unit.upper() == "M":
        multiplicator = 60
    elif origin_unit.upper() == "H":
        multiplicator = 3600
    else:
        print(origin_interval, origin_unit)
        raise NotImplementedError
    return origin_interval * multiplicator


def compute_throughput(interval: str, pkt_size_in_byte: int = 512) -> Union[float, int]:
    result = re.match(r"(\d+\.?\d*)(\w{1,2})", interval)
    return pkt_size_in_byte * 8 / convert_time_unit(float(result.group(1)),
                                                    result.group(2)) / 1000 if convert_time_unit(float(result.group(1)),
                                                                                                 result.group(
                                                                                                     2)) != 0 else 0


class AppStat:
    def __init__(self, pdr=None, adt=None, jitter=None, app_send=None, app_received=None, app_received_ooo=None):
        if app_received_ooo is None:
            app_received_ooo = []
        if app_received is None:
            app_received = []
        if app_send is None:
            app_send = []
        if jitter is None:
            jitter = []
        if pdr is None:
            pdr = []
        if adt is None:
            adt = []
        self.pdr = pdr
        self.adt = adt
        self.jitter = jitter
        self.app_send = app_send
        self.app_received = app_received
        self.app_received_ooo = app_received_ooo


class NetStat:
    def __init__(self, net_send=None, net_received=None, net_forward=None, drop_no_route=None, drop_expired_ttl=None,
                 drop_overflow=None):
        if drop_overflow is None:
            drop_overflow = list()
        if drop_expired_ttl is None:
            drop_expired_ttl = list()
        if drop_no_route is None:
            drop_no_route = list()
        if net_forward is None:
            net_forward = list()
        if net_send is None:
            net_send = list()
        if net_received is None:
            net_received = list()
        self.net_send = net_send
        self.net_received = net_received
        self.net_forward = net_forward
        self.drop_no_route = drop_no_route
        self.drop_expired_ttl = drop_expired_ttl
        self.drop_overflow = drop_overflow


class AppSetup:
    def __init__(self, app: str):
        self.app = app.replace("\\\\", "\\")  # type: str
        self.fg_throughput = self.get_mean_cbr_throughput()
        self.bg_throughput = self.get_mean_gen_traffic_throughput()

    def __eq__(self, other):
        if isinstance(other, AppSetup):
            return self.app == other.app
        return False

    def __hash__(self):
        return hash(self.app)

    def get_mean_cbr_throughput(self) -> Union[int, float]:
        cbr = self.app.replace("\r", "").split("\n")[:-1]
        avg = 0
        count = 0
        for x in cbr:
            tmp = x.split(" ")
            if x != "" and tmp[0] == "CBR":
                avg += compute_throughput(tmp[-3])
                count += 1
        if count > 0:
            return avg / count
        return 0

    def get_mean_gen_traffic_throughput(self) -> Union[int, float]:
        cbr = self.app.replace("\r", "").split("\n")[:-1]
        avg = 0
        count = 0
        for x in cbr:
            tmp = x.strip(" ").split(" ")
            if x != "" and tmp[0] == "TRAFFIC-GEN":
                avg += compute_throughput(tmp[-3])
                count += 1
        if count > 0:
            return avg / count
        return 0


class QroutingSetup:
    def __init__(self, sha_commit, alpha, exploration_phase, max_route_length, init_reward, timeout, periodic_update,
                 timeout_check_rate, nb_sample_for_latency, nb_sample_end_trial, epoch, epsilon, predicted_value,
                 extra):
        self.sha_commit = sha_commit
        self.alpha = alpha
        self.exploration_phase = exploration_phase
        self.max_route_length = max_route_length
        self.init_reward = init_reward
        self.timeout = timeout
        self.periodic_update = periodic_update
        self.timeout_check_rate = timeout_check_rate
        self.nb_sample_for_latency = nb_sample_for_latency
        self.nb_sample_end_trial = nb_sample_end_trial
        self.epoch = epoch
        self.epsilon = epsilon
        self.predicted_value = predicted_value
        self.extra = extra

    def to_string(self) -> str:
        return "_".join(map(str, (
            self.sha_commit, self.alpha, self.exploration_phase, self.max_route_length, self.init_reward, self.timeout,
            self.periodic_update, self.timeout_check_rate, self.nb_sample_for_latency, self.nb_sample_end_trial,
            self.epoch, self.epsilon, self.predicted_value, self.extra)))

    def __hash__(self):
        return hash(self.to_string())

    def __eq__(self, o) -> bool:
        if isinstance(o, QroutingSetup):
            return self.sha_commit == o.sha_commit and self.alpha == o.alpha and self.exploration_phase == o.exploration_phase and self.max_route_length == o.max_route_length and self.init_reward == o.init_reward and self.timeout == o.timeout and self.periodic_update == o.periodic_update and self.timeout_check_rate == o.timeout_check_rate and self.nb_sample_for_latency == o.nb_sample_for_latency and self.nb_sample_end_trial == o.nb_sample_end_trial and self.epoch == o.epoch and self.epsilon == o.epsilon and self.predicted_value == o.predicted_value and self.extra == o.extra
        return False


class Stats:
    def __init__(self, protocol: str, simulation_time: str, scenario: str):
        self.protocol = protocol  # type: str
        self.simulation_time = simulation_time  # type: str
        self.scenario = scenario  # type: str
        self.net_stat = None  # type: Union[NetStat, None]
        self.app_stat = None  # type: Union[AppStat, None]
        self.app = None  # type: Union[AppSetup, None]
        self.qs = None  # type: Union[QroutingSetup, None]

    def __eq__(self, other):
        if isinstance(other, Stats):
            return self.protocol == other.protocol and self.simulation_time == other.simulation_time and self.scenario == other.scenario and self.app == other.app and self.qs == other.qs
        else:
            return False

    def __hash__(self):
        return hash(self.protocol + self.simulation_time + self.scenario) ^ hash(self.app)

    def export_for_csv(self) -> List:
        def triple_stat(data: List[int, float]) -> tuple:
            return statistics.mean(data), statistics.median(data), statistics.stdev(data)

        return [self.protocol, self.simulation_time, self.app.fg_throughput, self.app.bg_throughput, self.qs.alpha,
                *triple_stat(self.app_stat.adt), *triple_stat(self.app_stat.pdr), *triple_stat(self.app_stat.jitter),
                statistics.mean(self.net_stat.drop_no_route), statistics.mean(self.net_stat.drop_expired_ttl),
                statistics.mean(self.net_stat.drop_overflow)]


def build_qrouting_setup(connect_to_db_func: Callable, stat: Stats, commit: str) -> List[Stats]:
    conn = connect_to_db_func()
    cur = conn.cursor()
    if commit != "":
        cur.execute(
            "SELECT DISTINCT `sha_commit`, `alpha`, `explorationPhase`, `maxRouteLength`, `initReward`, `timeout`, `periodicUpdate`, `timeoutCheckRate`, `nbSampleForLatency`, `nbSampleEndTrial`, `epoch`, `epsilon`, `predictedValue`, `extra` FROM qrouting NATURAL JOIN setup where setup.scenario like ? AND setup.duration like ? AND setup.protocol like ? AND qrouting.sha_commit=?;",
            (stat.scenario, stat.simulation_time, stat.protocol, commit))
        qs = cur.fetchall()
    else:
        cur.execute(
            "SELECT DISTINCT `sha_commit`, `alpha`, `explorationPhase`, `maxRouteLength`, `initReward`, `timeout`, `periodicUpdate`, `timeoutCheckRate`, `nbSampleForLatency`, `nbSampleEndTrial`, `epoch`, `epsilon`, `predictedValue`, `extra` FROM qrouting NATURAL JOIN setup where setup.scenario like ? AND setup.duration like ? AND setup.protocol like ?;",
            (stat.scenario, stat.simulation_time, stat.protocol))
        qs = cur.fetchall()
    cur.close()
    conn.close()
    data = list()  # type: List[Stats]
    for elm in qs:
        tmp = QroutingSetup(*elm)
        tmp_stat = copy.deepcopy(stat)
        tmp_stat.qs = tmp
        data.append(tmp_stat)
    return data


def build_app(connect_to_db_func: Callable, stat: Stats) -> List[Stats]:
    conn = connect_to_db_func()
    cur = conn.cursor()
    if stat.qs is not None:
        cur.execute(
            "SELECT DISTINCT `app` FROM `app` NATURAL JOIN setup NATURAL JOIN qrouting where setup.scenario like ? AND setup.duration like ? AND setup.protocol like ? AND `sha_commit`=? AND `alpha`=? AND `explorationPhase`=? AND `maxRouteLength`=? AND `initReward`=? AND `timeout`=? AND `periodicUpdate`=? AND `timeoutCheckRate`=? AND `nbSampleForLatency`=? AND `nbSampleEndTrial`=? AND `epoch`=? AND `epsilon`=? AND `predictedValue`=? AND `extra`=?;",
            (stat.scenario, stat.simulation_time, stat.protocol, stat.qs.sha_commit, stat.qs.alpha,
             stat.qs.exploration_phase, stat.qs.max_route_length, stat.qs.init_reward, stat.qs.timeout,
             stat.qs.periodic_update, stat.qs.timeout_check_rate, stat.qs.nb_sample_for_latency,
             stat.qs.nb_sample_end_trial, stat.qs.epoch, stat.qs.epsilon, stat.qs.predicted_value,
             stat.qs.extra))
        ns = cur.fetchall()
    else:
        cur.execute(
            "SELECT DISTINCT `app` FROM `app` NATURAL JOIN setup where setup.scenario like ? AND setup.duration like ? AND setup.protocol like ?;",
            (stat.scenario, stat.simulation_time, stat.protocol))
        ns = cur.fetchall()
    cur.close()
    conn.close()
    data = list()
    for elm in ns:
        tmp = AppSetup(*elm)
        stat_copy = copy.deepcopy(stat)
        stat_copy.app = tmp
        data.append(stat_copy)
    return data


def build_net_stat(connect_to_db_func: Callable, stat: Stats):
    conn = connect_to_db_func()
    cur = conn.cursor()
    if stat.qs is not None:
        cur.execute(
            "SELECT `net_send`, `net_recieve`, `net_forward`, `drop_no_route`, `drop_expired_ttl`, `drop_overflow` FROM `stats_net` NATURAL JOIN setup NATURAL JOIN qrouting NATURAL JOIN app where setup.scenario like ? AND setup.duration like ? AND setup.protocol like ? AND `sha_commit`=? AND `alpha`=? AND `explorationPhase`=? AND `maxRouteLength`=? AND `initReward`=? AND `timeout`=? AND `periodicUpdate`=? AND `timeoutCheckRate`=? AND `nbSampleForLatency`=? AND `nbSampleEndTrial`=? AND `epoch`=? AND `epsilon`=? AND `predictedValue`=? AND `extra`=? AND app.app=?;",
            (stat.scenario, stat.simulation_time, stat.protocol, stat.qs.sha_commit, stat.qs.alpha,
             stat.qs.exploration_phase, stat.qs.max_route_length, stat.qs.init_reward, stat.qs.timeout,
             stat.qs.periodic_update, stat.qs.timeout_check_rate, stat.qs.nb_sample_for_latency,
             stat.qs.nb_sample_end_trial, stat.qs.epoch, stat.qs.epsilon, stat.qs.predicted_value, stat.qs.extra,
             stat.app.app))
        ns = cur.fetchall()
    else:
        cur.execute(
            "SELECT `net_send`, `net_recieve`, `net_forward`, `drop_no_route`, `drop_expired_ttl`, `drop_overflow` FROM `stats_net` NATURAL JOIN setup NATURAL JOIN app where setup.scenario like ? AND setup.duration like ? AND setup.protocol like ? AND app.app=?;",
            (stat.scenario, stat.simulation_time, stat.protocol, stat.app.app))
        ns = cur.fetchall()
    cur.close()
    conn.close()
    tmp = NetStat()
    for elm in ns:
        tmp.net_send.append(elm[0])
        tmp.net_received.append(elm[1])
        tmp.net_forward.append(elm[2])
        tmp.drop_no_route.append(elm[3])
        tmp.drop_expired_ttl.append(elm[4])
        tmp.drop_overflow.append(elm[5])
    stat.net_stat = tmp


def build_app_stat(connect_to_db_func: Callable, stat: Stats):
    conn = connect_to_db_func()
    cur = conn.cursor()
    if stat.qs is not None:
        cur.execute(
            "SELECT `delay`, `pdr`, `jitter`, `app_send`, `app_recieve`, `app_recieve_ooo` FROM `stats_app` NATURAL JOIN setup NATURAL JOIN qrouting NATURAL JOIN app where setup.scenario like ? AND setup.duration like ? AND setup.protocol like ? AND `sha_commit`=? AND `alpha`=? AND `explorationPhase`=? AND `maxRouteLength`=? AND `initReward`=? AND `timeout`=? AND `periodicUpdate`=? AND `timeoutCheckRate`=? AND `nbSampleForLatency`=? AND `nbSampleEndTrial`=? AND `epoch`=? AND `epsilon`=? AND `predictedValue`=? AND `extra`=? AND app.app=?;",
            (stat.scenario, stat.simulation_time, stat.protocol, stat.qs.sha_commit, stat.qs.alpha,
             stat.qs.exploration_phase, stat.qs.max_route_length, stat.qs.init_reward, stat.qs.timeout,
             stat.qs.periodic_update, stat.qs.timeout_check_rate, stat.qs.nb_sample_for_latency,
             stat.qs.nb_sample_end_trial, stat.qs.epoch, stat.qs.epsilon, stat.qs.predicted_value, stat.qs.extra,
             stat.app.app))
        ns = cur.fetchall()
    else:
        cur.execute(
            "SELECT `delay`, `pdr`, `jitter`, `app_send`, `app_recieve`, `app_recieve_ooo` FROM `stats_app` NATURAL JOIN setup NATURAL JOIN app where setup.scenario like ? AND setup.duration like ? AND setup.protocol like ? AND app.app=?;",
            (stat.scenario, stat.simulation_time, stat.protocol, stat.app.app))
        ns = cur.fetchall()
    cur.close()
    conn.close()
    tmp = AppStat()
    for elm in ns:
        tmp.adt.append(elm[0])
        tmp.pdr.append(elm[1])
        tmp.jitter.append(elm[2])
        tmp.app_send.append(elm[3])
        tmp.app_received.append(elm[4])
        tmp.app_received_ooo.append(elm[5])
    stat.app_stat = tmp


def __make_extra_param(string: str, key: str, value: str) -> str:
    if value != "":
        if string != "":
            string += " and "
        string += " %s='%s' " % (key, value)
    return string


def build_line_stat_from_sqlite(connect_to_db_func: Callable, scenario: str = "", protocol: str = "",
                                duration: str = "",
                                commit: str = "") -> List[Stats]:
    print("Building stats")
    data = list()  # type: List[Stats]
    conn = connect_to_db_func()
    cur = conn.cursor()
    extra_param = __make_extra_param("", "scenario", scenario)
    extra_param = __make_extra_param(extra_param, "protocol", protocol)
    extra_param = __make_extra_param(extra_param, "duration", duration)
    if extra_param != "":
        cur.execute("SELECT DISTINCT protocol, duration, scenario from setup where " + extra_param)
        setup = cur.fetchall()
    else:
        setup = cur.execute("SELECT DISTINCT protocol, duration, scenario from setup ;").fetchall()
    cur.close()
    conn.close()
    for elm in setup:
        stat = Stats(*elm)
        if stat.protocol == "QROUTING":
            print("Building Q-routing stats")
            tmp = build_qrouting_setup(connect_to_db_func, stat, commit)
            for qs in tmp:
                data += build_app(connect_to_db_func, qs)
        else:
            data += build_app(connect_to_db_func, stat)
    print("Building app and net stats")
    for stat in data:
        # build_net_stat(filename, stat)
        build_app_stat(connect_to_db_func, stat)
    return data


def group_stat_by_protocol(data: List[Stats]) -> dict:
    result = dict()
    for elm in data:
        if elm.protocol != "QROUTING":
            key = elm.protocol
        else:
            key = elm.protocol + elm.qs.to_string()
        key = key.replace("\\", "")
        if key in result.keys():
            result[key].append(elm)
        else:
            result[key] = [elm]
    return result


def filter_by_scenario(input_data: List[Stats], scenario: str) -> List[Stats]:
    return list(filter(lambda x: x.scenario == scenario, input_data))


def filter_by_protocol(input_data: List[Stats], protocol: str) -> List[Stats]:
    return list(filter(lambda x: x.protocol == protocol, input_data))
