#!/usr/bin/env python3

"""
Copyright or © or Copr. University of Nantes / LS2N (2021)

Contributor:
* Alexis Bitaillou

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
"""

from re import match

import svgwrite


class App:
    def __init__(self, source, dest, start, stop):
        self.source = source
        self.dest = dest
        self.start = [start]
        self.stop = [stop]

    def __eq__(self, other):
        if isinstance(other, App):
            return self.dest == other.dest and self.source == self.source
        return False

    def __hash__(self):
        return hash(self.dest) ^ hash(self.start)

    def add_start_stop(self, start, stop):
        self.start.append(start)
        self.stop.append(stop)


def convert_time(original_value: str) -> float:
    result = match(r"(\d+)(\w+)", original_value)
    if result is not None:
        origin_unit = result.group(2)
        if origin_unit.upper() == "MS":
            multiplicator = 1 / 1000
        elif origin_unit.upper() == "US":
            multiplicator = 1 / 1000000
        elif origin_unit.upper() == "S":
            multiplicator = 1
        elif origin_unit.upper() == "M":
            multiplicator = 60
        elif origin_unit.upper() == "H":
            multiplicator = 3600
        else:
            raise NotImplementedError
        return int(result.group(1)) * multiplicator


def parse_app_file(filename: str):
    # CBR 5 22 1000000000 512 800US 60S 29M
    with open(filename, 'r', encoding='utf8') as file:
        data = file.read()
    fsplit = data.split("\n")
    app_list = list()
    for elm in fsplit:
        if elm != "":
            tmp = elm.rstrip().split(" ")
            app = App(tmp[1], tmp[2], convert_time(tmp[-2]), convert_time(tmp[-1]))
            if app in app_list:
                for ap in app_list:
                    if app == ap:
                        ap.add_start_stop(convert_time(tmp[-2]), convert_time(tmp[-1]))
                        break
            else:
                app_list.append(app)
    return app_list


def app_drawer_factory(app: App, dwg: svgwrite.Drawing, multiplicator: int, offset: int):
    source = str(app.source)
    if len(source) < 2:
        source += " "
    dest = str(app.dest)
    if len(dest) < 2:
        dest += " "
    dwg.add(
        dwg.text("(%s, %s)" % (source, dest), insert=(5, 15 * multiplicator + 5), stroke='none',
                 fill=svgwrite.rgb(15, 15, 15, '%'),
                 font_size='15px', font_family="monospace"))
    for start, stop in zip(app.start, app.stop):
        dwg.add(dwg.line((start / 5 + offset, 15 * multiplicator), (stop / 5 + offset, 15 * multiplicator),
                         stroke="#000000", stroke_width=1.3, stroke_linecap="butt"))


def draw_axis(max_value: int, step: int, y_value: int, dwg: svgwrite.Drawing, offset):
    dwg.add(dwg.line((offset, y_value), (max_value + offset + step, y_value), stroke="#000000",
                     stroke_width=1.5, stroke_linecap="butt"))
    for i in range(offset + int(step / 2), max_value + offset + step, step):
        dwg.add(dwg.line((i, -10), (i, y_value), stroke="#b3b3b3",
                         stroke_width=0.9, stroke_linecap="butt"))
    for i in range(offset, max_value + offset + step, step):
        dwg.add(dwg.line((i, -10), (i, y_value), stroke="#999999",
                         stroke_width=0.9, stroke_linecap="butt"))
        dwg.add(dwg.line((i, y_value - 5), (i, y_value + 5), stroke="#000000",
                         stroke_width=0.9, stroke_linecap="butt"))
        dwg.add(dwg.text(str(int(((i - offset) * 5) / 60)), insert=(i - 4, y_value + 13), stroke='none',
                         fill=svgwrite.rgb(15, 15, 15, '%'), font_size='12px'))


def wireless_grid():
    dwg = svgwrite.Drawing(filename="wg2.svg", profile='tiny')
    tmp = parse_app_file("wireless_grid_2.app")
    tmp = sorted(tmp, key=lambda x: x.start)
    draw_axis(int(max(tmp, key=lambda x: x.stop).stop[-1] / 2), 50, (len(tmp) + 1) * 15, dwg, 100)
    for i in range(len(tmp)):
        app_drawer_factory(tmp[i], dwg, i, 100)
    dwg.save()


def grid94():
    dwg = svgwrite.Drawing(filename="grid94.svg", profile='tiny')
    tmp = parse_app_file("grid94.app")
    tmp = sorted(tmp, key=lambda x: x.start)
    draw_axis(int(max(tmp, key=lambda x: x.stop).stop[-1] / 5), 60, (len(tmp) + 1) * 15, dwg, 100)
    for i in range(len(tmp)):
        app_drawer_factory(tmp[i], dwg, i, 100)
    dwg.save()


if __name__ == "__main__":
    grid94()
