#!/usr/bin/env python3

"""
Copyright or © or Copr. University of Nantes / LS2N (2021)

Contributor:
* Alexis Bitaillou

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
"""


import os
import re
import statistics
from typing import List, Union, Callable

import matplotlib.pyplot as plt
from matplotlib.container import ErrorbarContainer

import db_connector
from analyse_exp_sqlite import build_line_stat_from_sqlite, group_stat_by_protocol


def convert_time_unit(origin_interval: float, origin_unit: str) -> float:
    if origin_unit.upper() == "MS":
        multiplicator = 1 / 1000
    elif origin_unit.upper() == "US":
        multiplicator = 1 / 1000000
    elif origin_unit.upper() == "S":
        multiplicator = 1
    elif origin_unit.upper() == "M":
        multiplicator = 60
    elif origin_unit.upper() == "H":
        multiplicator = 3600
    else:
        print(origin_interval)
        print(origin_unit)
        raise NotImplementedError
    return origin_interval * multiplicator


def compute_throughput(interval: str, pkt_size_in_byte: int = 512) -> float:
    result = re.match("(\d+\.?\d*)(\w{1,2})", interval)
    return (pkt_size_in_byte * 8 / convert_time_unit(float(result.group(1)), result.group(2))) / 1000


def pretty_print(interval: str, pkt_size_in_byte: int = 512) -> int:
    result = re.match("(\d+\.?\d*)(\w{1,2})", interval)
    return int("%.0f" % ((pkt_size_in_byte * 8 / convert_time_unit(float(result.group(1)), result.group(2))) / 1000))


class State:
    def __init__(self, connect_to_db_func: Callable, second_proto: List[str], scenario: str, simulation_time: str,
                 commit: str = ""):
        self.raw_data = sorted(
            build_line_stat_from_sqlite(connect_to_db_func, scenario, duration=simulation_time, protocol="QROUTING",
                                        commit=commit),
            key=lambda stat: stat.app.fg_throughput)
        self.qrouting = group_stat_by_protocol(self.raw_data)
        self.color = ["red", "blue", "green", "orange", "purple", "cyan", "yellow"]
        self.linestyle = ["-", "--", ":", '-.', (0, (5, 10)), (0, (5, 1))]
        self.second_proto = second_proto  # type: List[str]
        self.output_base_file = "%s_%s" % (scenario, simulation_time)
        self.second_proto_data = dict()  # type: dict
        for proto in self.second_proto:
            self.second_proto_data[proto] = sorted(
                build_line_stat_from_sqlite(connect_to_db_func, scenario=scenario, protocol=proto,
                                            duration=simulation_time),
                key=lambda stat: stat.app.fg_throughput)

    def plot_qrouting_mean_error_metric(self, metric_x: str, metric_y: str, key, axis,
                                        color_index: int, scale_x: Union[float, int] = 1,
                                        scale_y: Union[float, int] = 1) -> ErrorbarContainer:
        pos = [(x.app.__getattribute__(metric_x) / scale_x) for x in self.qrouting[key]]
        mpdr = [statistics.mean(x.app_stat.__getattribute__(metric_y)) / scale_y for x in self.qrouting[key]]
        sd_pdr = [statistics.stdev(x.app_stat.__getattribute__(metric_y)) / scale_y for x in self.qrouting[key]]
        print(metric_y + " First: %s / Last: %s" % (mpdr[0], mpdr[-1]))
        lh = axis.errorbar(pos, mpdr, yerr=sd_pdr, color=self.color[color_index % len(self.color)],
                           linestyle=self.linestyle[color_index % len(self.linestyle)], marker='x')
        return lh

    def plot_sec_mean_error_metric(self, metric_x: str, metric_y: str, protocol: str, axis, color_index: int,
                                   scale_x: Union[float, int] = 1, scale_y: Union[float, int] = 1) -> ErrorbarContainer:
        pos = [(x.app.__getattribute__(metric_x) / scale_x) for x in self.second_proto_data[protocol]]
        mpdr = [statistics.mean(x.app_stat.__getattribute__(metric_y)) / scale_y for x in
                self.second_proto_data[protocol]]
        sd_pdr = [statistics.stdev(x.app_stat.__getattribute__(metric_y)) / scale_y for x in
                  self.second_proto_data[protocol]]
        lh = axis.errorbar(pos, mpdr, yerr=sd_pdr, color=self.color[color_index % len(self.color)],
                           linestyle=self.linestyle[color_index % len(self.linestyle)], marker='x')
        return lh

    def plot_mean_error_pdr(self, key: str, axis, color):
        return self.plot_qrouting_mean_error_metric("fg_throughput", "pdr", key, axis, color, 1000, 0.01)

    def plot_mean_error_adt(self, key, axis, color):
        return self.plot_qrouting_mean_error_metric("fg_throughput", "adt", key, axis, color, 1000, 0.001)

    def plot_mean_error_pdr_sec(self, protocol: str, axis, color: int):
        return self.plot_sec_mean_error_metric("fg_throughput", "pdr", protocol, axis, color, 1000, 0.01)

    def plot_mean_error_adt_sec(self, protocol: str, axis, color: int):
        return self.plot_sec_mean_error_metric("fg_throughput", "adt", protocol, axis, color, 1000, 0.001)

    def plot_mean_error_jitter(self, key, axis, color):
        return self.plot_qrouting_mean_error_metric("fg_throughput", "jitter", key, axis, color, 1000,
                                                    0.001)

    def plot_mean_error_jitter_sec(self, protocol: str, axis, color: int):
        return self.plot_sec_mean_error_metric("fg_throughput", "jitter", protocol, axis, color, 1000, 0.001)


def grid94():
    def wrapper_db():
        return db_connector.connect_to_mariadb_db("../Test/db.json")
        # return db_connector.connect_to_sqlite_db("results.db", None)

    state = State(wrapper_db, ["OLSRv2-NIIGATA"], "grid94", "61M", "adhocnet2020-original_7efd6c")
    print(state.qrouting.keys())

    qr = ('Q-routing', 'QROUTINGadhocnet2020-original_7efd6c0_0.9_15_16_0_60_10_60_0_0_0_0_0_{}', 0)
    proto = [qr]
    sproto = [('nuOLSRv2', 'OLSRv2-NIIGATA', 1)]
    return state, proto, sproto, "grid94"


def wg2():
    def wrapper_db():
        return db_connector.connect_to_mariadb_db("../Test/db.json")
        # return db_connector.connect_to_sqlite_db("results.db", None)

    state = State(wrapper_db, ["OLSRv2-NIIGATA"], "wireless_grid_2", "30M", "q-routing_84caa2e")
    print(state.qrouting.keys())

    qr = ('Q-routing', 'QROUTINGq-routing_84caa2e_0.9_45_16_0_60_10_60_7_3_0_0_1_{}', 0)
    proto = [qr]
    sproto = [('nuOLSRv2', 'OLSRv2-NIIGATA', 1)]
    return state, proto, sproto, "wg2"


if __name__ == '__main__':
    state1, proto, sproto, grid = grid94()
    fig1, ax1 = plt.subplots()
    lh1 = dict()

    fig2, ax2 = plt.subplots()
    lh2 = dict()

    fig3, ax3 = plt.subplots()
    lh3 = dict()

    for x in proto:
        lh1[x[0]] = state1.plot_mean_error_pdr(x[1], ax1, x[2])
        lh2[x[0]] = state1.plot_mean_error_adt(x[1], ax2, x[2])
        lh3[x[0]] = state1.plot_mean_error_jitter(x[1], ax3, x[2])
    for x in sproto:
        lh1[x[0]] = state1.plot_mean_error_pdr_sec(x[1], ax1, x[2])
        lh2[x[0]] = state1.plot_mean_error_adt_sec(x[1], ax2, x[2])
        lh3[x[0]] = state1.plot_mean_error_jitter_sec(x[1], ax3, x[2])

    for ax, lh in zip((ax1, ax2, ax3), (lh1, lh2, lh3)):
        ax.legend((lh.values()), (lh.keys()), loc="best")
        ax.grid()

    ax1.set(xlabel='Débit par flux (Mb/s)', ylabel='Taux de paquets livrés (%)')
    ax2.set(xlabel='Débit par flux (Mb/s)', ylabel='Délai de livraison moyen (ms)')
    ax3.set(xlabel='Débit par flux (Mb/s)', ylabel='Gigue moyenne (ms)')

    for f, m in zip((fig1, fig2, fig3), ("pdr", "adt", "jitter")):
        f.tight_layout()
        f.savefig(str(grid) + "_" + m + ".pdf", transparent=True)
    plt.close()
