#!/usr/bin/env python3

"""
Copyright or © or Copr. University of Nantes / LS2N (2021)

Contributor:
* Alexis Bitaillou

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
"""

import copy
import re
import statistics
from typing import List, Union, Callable

import matplotlib.pyplot as plt
from matplotlib.container import ErrorbarContainer

import db_connector
from analyse_exp_sqlite import group_stat_by_protocol, filter_by_protocol, AppSetup, \
    QroutingSetup, NetStat, AppStat


def convert_time_unit(origin_interval: float, origin_unit: str) -> float:
    if origin_unit.upper() == "MS":
        multiplicator = 1 / 1000
    elif origin_unit.upper() == "US":
        multiplicator = 1 / 1000000
    elif origin_unit.upper() == "S":
        multiplicator = 1
    elif origin_unit.upper() == "M":
        multiplicator = 60
    elif origin_unit.upper() == "H":
        multiplicator = 3600
    else:
        print(origin_interval)
        print(origin_unit)
        raise NotImplementedError
    return origin_interval * multiplicator


def compute_throughput(interval: str, pkt_size_in_byte: int = 512) -> float:
    result = re.match("(\d+\.?\d*)(\w{1,2})", interval)
    return (pkt_size_in_byte * 8 / convert_time_unit(float(result.group(1)), result.group(2))) / 1000


def pretty_print(interval: str, pkt_size_in_byte: int = 512) -> int:
    result = re.match("(\d+\.?\d*)(\w{1,2})", interval)
    return int("%.0f" % ((pkt_size_in_byte * 8 / convert_time_unit(float(result.group(1)), result.group(2))) / 1000))


def convert_speed(simulation_time: str):
    return 1300 / ((int(simulation_time.replace("M", "")) - 2) * 60 - 35)


class Stats:
    def __init__(self, protocol: str, simulation_time: List[str], scenario: str):
        self.protocol = protocol  # type: str
        self.simulation_time = simulation_time  # type: List[str]
        self.scenario = scenario  # type: str
        self.net_stat = None  # type: Union[NetStat, None]
        self.app_stat = None  # type: Union[AppStat, None]
        self.app = dict()  # type: dict
        self.qs = None  # type: Union[QroutingSetup, None]

    def __eq__(self, other):
        if isinstance(other, Stats):
            return self.protocol == other.protocol and self.scenario == other.scenario and self.app == other.app and self.qs == other.qs
        else:
            return False

    def __hash__(self):
        return hash(self.protocol + self.scenario) ^ hash(self.app)


class MobilAppSetup(AppSetup):
    def __init__(self, app: str, simulation_time: str):
        super().__init__(app)
        self.net_stat = None  # type: Union[NetStat, None]
        self.app_stat = None  # type: Union[AppStat, None]
        self.speed = convert_speed(simulation_time)


def build_app(connect_db_func: Callable, stat: Stats, simulation_time: str) -> List[Stats]:
    conn = connect_db_func()
    cursor = conn.cursor()
    if stat.qs is not None:
        ns = cursor.execute(
            "SELECT DISTINCT `app` FROM `app` NATURAL JOIN setup NATURAL JOIN qrouting where setup.scenario like ? AND setup.duration like ? AND setup.protocol like ? AND `sha_commit`=? AND `alpha`=? AND `explorationPhase`=? AND `maxRouteLength`=? AND `initReward`=? AND `timeout`=? AND `periodicUpdate`=? AND `timeoutCheckRate`=? AND `nbSampleForLatency`=? AND `nbSampleEndTrial`=? AND `epoch`=? AND `epsilon`=? AND `predictedValue`=? AND `extra`=?;",
            (stat.scenario, simulation_time, stat.protocol, stat.qs.sha_commit, stat.qs.alpha,
             stat.qs.exploration_phase, stat.qs.max_route_length, stat.qs.init_reward, stat.qs.timeout,
             stat.qs.periodic_update, stat.qs.timeout_check_rate, stat.qs.nb_sample_for_latency,
             stat.qs.nb_sample_end_trial, stat.qs.epoch, stat.qs.epsilon, stat.qs.predicted_value,
             stat.qs.extra)).fetchall()
    else:
        ns = conn.execute(
            "SELECT DISTINCT `app` FROM `app` NATURAL JOIN setup where setup.scenario like ? AND setup.duration like ? AND setup.protocol like ?;",
            (stat.scenario, simulation_time, stat.protocol)).fetchall()
    data = list()
    for elm in ns:
        tmp = MobilAppSetup(*elm, simulation_time)
        stat.app[simulation_time] = tmp
    cursor.close()
    conn.close()
    return data


def build_qrouting_setup(connect_db_func: Callable, stat: Stats) -> List[Stats]:
    conn = connect_db_func()
    cursor = conn.cursor()
    qs = cursor.execute(
        "SELECT DISTINCT `sha_commit`, `alpha`, `explorationPhase`, `maxRouteLength`, `initReward`, `timeout`, `periodicUpdate`, `timeoutCheckRate`, `nbSampleForLatency`, `nbSampleEndTrial`, `epoch`, `epsilon`, `predictedValue`, `extra` FROM qrouting NATURAL JOIN setup where setup.scenario like ? AND setup.protocol like ?;",
        (stat.scenario, stat.protocol)).fetchall()
    data = list()  # type: List[Stats]
    for elm in qs:
        tmp = QroutingSetup(*elm)
        tmp_stat = copy.deepcopy(stat)
        tmp_stat.qs = tmp
        data.append(tmp_stat)
    cursor.close()
    conn.close()
    return data


def build_net_stat(connect_db_func: Callable, stat: Stats, simulation_time: str):
    conn = connect_db_func()
    cursor = conn.cursor()
    if stat.qs is not None:
        ns = cursor.execute(
            "SELECT `net_send`, `net_recieve`, `net_forward`, `drop_no_route`, `drop_expired_ttl`, `drop_overflow` FROM `stats_net` NATURAL JOIN setup NATURAL JOIN qrouting NATURAL JOIN app where setup.scenario like ? AND setup.protocol like ? AND `sha_commit`=? AND `alpha`=? AND `explorationPhase`=? AND `maxRouteLength`=? AND `initReward`=? AND `timeout`=? AND `periodicUpdate`=? AND `timeoutCheckRate`=? AND `nbSampleForLatency`=? AND `nbSampleEndTrial`=? AND `epoch`=? AND `epsilon`=? AND `predictedValue`=? AND `extra`=? AND app.app=?;",
            (stat.scenario, stat.protocol, stat.qs.sha_commit, stat.qs.alpha,
             stat.qs.exploration_phase, stat.qs.max_route_length, stat.qs.init_reward, stat.qs.timeout,
             stat.qs.periodic_update, stat.qs.timeout_check_rate, stat.qs.nb_sample_for_latency,
             stat.qs.nb_sample_end_trial, stat.qs.epoch, stat.qs.epsilon, stat.qs.predicted_value, stat.qs.extra,
             stat.app[simulation_time].app)).fetchall()
    else:
        ns = cursor.execute(
            "SELECT `net_send`, `net_recieve`, `net_forward`, `drop_no_route`, `drop_expired_ttl`, `drop_overflow` FROM `stats_net` NATURAL JOIN setup NATURAL JOIN app where setup.scenario like ? AND setup.protocol like ? AND app.app=?;",
            (stat.scenario, stat.protocol, stat.app[simulation_time].app)).fetchall()
    tmp = NetStat()
    for elm in ns:
        tmp.net_send.append(elm[0])
        tmp.net_received.append(elm[1])
        tmp.net_forward.append(elm[2])
        tmp.drop_no_route.append(elm[3])
        tmp.drop_expired_ttl.append(elm[4])
        tmp.drop_overflow.append(elm[5])
    stat.app[simulation_time].net_stat = tmp
    cursor.close()
    conn.close()


def build_app_stat(connect_db_func: Callable, stat: Stats, simulation_time: str):
    conn = connect_db_func()
    cursor = conn.cursor()
    if stat.qs is not None:
        ns = cursor.execute(
            "SELECT `delay`, `pdr`, `jitter`, `app_send`, `app_recieve`, `app_recieve_ooo` FROM `stats_app` NATURAL JOIN setup NATURAL JOIN qrouting NATURAL JOIN app where setup.scenario like ? AND setup.protocol like ? AND `sha_commit`=? AND `alpha`=? AND `explorationPhase`=? AND `maxRouteLength`=? AND `initReward`=? AND `timeout`=? AND `periodicUpdate`=? AND `timeoutCheckRate`=? AND `nbSampleForLatency`=? AND `nbSampleEndTrial`=? AND `epoch`=? AND `epsilon`=? AND `predictedValue`=? AND `extra`=? AND app.app=?;",
            (stat.scenario, stat.protocol, stat.qs.sha_commit, stat.qs.alpha,
             stat.qs.exploration_phase, stat.qs.max_route_length, stat.qs.init_reward, stat.qs.timeout,
             stat.qs.periodic_update, stat.qs.timeout_check_rate, stat.qs.nb_sample_for_latency,
             stat.qs.nb_sample_end_trial, stat.qs.epoch, stat.qs.epsilon, stat.qs.predicted_value, stat.qs.extra,
             stat.app[simulation_time].app)).fetchall()
    else:
        ns = cursor.execute(
            "SELECT `delay`, `pdr`, `jitter`, `app_send`, `app_recieve`, `app_recieve_ooo` FROM `stats_app` NATURAL JOIN setup NATURAL JOIN app where setup.scenario like ? AND setup.protocol like ? AND app.app=?;",
            (stat.scenario, stat.protocol, stat.app[simulation_time].app)).fetchall()
    tmp = AppStat()
    for elm in ns:
        tmp.adt.append(elm[0])
        tmp.pdr.append(elm[1])
        tmp.jitter.append(elm[2])
        tmp.app_send.append(elm[3])
        tmp.app_received.append(elm[4])
        tmp.app_received_ooo.append(elm[5])
    stat.app[simulation_time].app_stat = tmp
    cursor.close()
    conn.close()


def build_line_stat_from_sqlite(connect_db, scenario: str = "") -> List[Stats]:
    print("Building stats")
    data = list()  # type: List[Stats]
    conn = connect_db()
    cursor = conn.cursor()
    if scenario != "":
        setup = cursor.execute("SELECT DISTINCT protocol, duration, scenario from setup where scenario=?;",
                             (scenario,)).fetchall()
    else:
        setup = cursor.execute("SELECT DISTINCT protocol, duration, scenario from setup ;").fetchall()
    for elm in setup:
        found = False
        stat = Stats(elm[0], [elm[1]], elm[2])
        for elm1 in data:
            if stat == elm1:
                elm1.simulation_time.append(elm[1])
                found = True
        if not found:
            data.append(stat)

    counter = 1
    data_length = len(data)
    stat_to_remove = list()
    qrouting_stat = list()
    for stat in data:
        stat.simulation_time = sorted(stat.simulation_time, key=lambda x: convert_speed(x))
        if stat.protocol == "QROUTING":
            print("[%s/%s] Building Q-routing stats" % (counter, data_length))
            tmp = build_qrouting_setup(connect_db, stat)
            qrouting_stat += tmp
            counter += 1
            stat_to_remove.append(stat)
    for elm in stat_to_remove:
        data.remove(elm)
    data += qrouting_stat
    counter = 1
    for stat in data:
        print("[%s/%s] Building app and net stats" % (counter, len(data)))
        for st in stat.simulation_time:
            build_app(connect_db, stat, st)
            build_net_stat(connect_db, stat, st)
            build_app_stat(connect_db, stat, st)
        counter += 1
    cursor.close()
    conn.close()
    return data


class State:
    def __init__(self, connect_db_func: Callable, second_proto: List[str], scenario: str):
        self.raw_data = build_line_stat_from_sqlite(connect_db_func, scenario)  # type: List[Stats]
        self.qrouting = group_stat_by_protocol(filter_by_protocol(self.raw_data, "QROUTING"))
        self.color = ["red", "blue", "green", "orange", "purple", "cyan", "yellow"]
        self.linestyle = ["-", "--", ":", '-.', (0, (5, 10)), (0, (5, 1))]
        self.second_proto = second_proto  # type: List[str]
        self.output_base_file = "%s" % scenario
        self.second_proto_data = dict()  # type: dict
        for proto in self.second_proto:
            self.second_proto_data[proto] = filter_by_protocol(self.raw_data, proto)

    def plot_qrouting_mean_error_metric(self, metric_x: str, metric_y: str, key, axis,
                                        color_index: int, scale_x: Union[float, int] = 1,
                                        scale_y: Union[float, int] = 1) -> ErrorbarContainer:
        pos = [(self.qrouting[key][0].app[x].__getattribute__(metric_x) / scale_x) for x in
               self.qrouting[key][0].simulation_time]
        mpdr = [statistics.mean(self.qrouting[key][0].app[x].app_stat.__getattribute__(metric_y)) / scale_y for x in
                self.qrouting[key][0].simulation_time]
        sd_pdr = [statistics.stdev(self.qrouting[key][0].app[x].app_stat.__getattribute__(metric_y)) / scale_y for x in
                  self.qrouting[key][0].simulation_time]
        lh = axis.errorbar(pos, mpdr, yerr=sd_pdr, color=self.color[color_index], linestyle=self.linestyle[color_index])
        return lh

    def plot_sec_mean_error_metric(self, metric_x: str, metric_y: str, protocol: str, axis, color_index: int,
                                   scale_x: Union[float, int] = 1, scale_y: Union[float, int] = 1) -> ErrorbarContainer:
        pos = [(self.second_proto_data[protocol].app[x].__getattribute__(metric_x) / scale_x) for x in
               self.second_proto_data[protocol][0].simulation_time]
        mpdr = [statistics.mean(self.second_proto_data[protocol].app[x].app_stat.__getattribute__(metric_y)) / scale_y
                for x in
                self.second_proto_data[protocol]]
        sd_pdr = [
            statistics.stdev(self.second_proto_data[protocol].app[x].app_stat.__getattribute__(metric_y)) / scale_y for
            x in
            self.second_proto_data[protocol].simulation_time]
        lh = axis.errorbar(pos, mpdr, yerr=sd_pdr, color=self.color[color_index], linestyle=self.linestyle[color_index])
        return lh

    def plot_mean_error_pdr(self, key: str, ax, color):
        return self.plot_qrouting_mean_error_metric("speed", "pdr", key, ax, color, 1, 0.01)

    def plot_mean_error_adt(self, key, ax, color):
        return self.plot_qrouting_mean_error_metric("speed", "adt", key, ax, color, 1, 0.001)

    def plot_mean_error_pdr_sec(self, protocol: str, ax, color: int):
        return self.plot_sec_mean_error_metric("speed", "pdr", protocol, ax, color, 1, 0.01)

    def plot_mean_error_adt_sec(self, protocol: str, ax, color: int):
        return self.plot_sec_mean_error_metric("speed", "adt", protocol, ax, color, 1, 0.001)

    def plot_mean_error_jitter(self, key, ax, color):
        return self.plot_qrouting_mean_error_metric("speed", "jitter", key, ax, color, 1,
                                                    0.001)

    def plot_mean_error_jitter_sec(self, protocol: str, ax, color: int):
        return self.plot_sec_mean_error_metric("speed", "jitter", protocol, ax, color, 1, 0.001)


def hipo():
    def wrapper_db():
        # return db_connector.connect_to_mariadb_db("../Test/db.json")
        return db_connector.connect_to_sqlite_db("results.db", None)
    state = State(wrapper_db, ["OLSRv2-NIIGATA"], "hipo")
    print(state.qrouting.keys())

    qr = ('Q-routing', 'QROUTINGadhocnet2020-original_7efd6c0_0.9_15_16_0_10_5_10_0_0_0_0_0_{}', 0)
    proto = [qr]
    sproto = [('nuOLSRv2', 'OLSRv2-NIIGATA', 1)]
    return state, proto, sproto, "hipo"


if __name__ == '__main__':
    state1, proto, sproto, grid = hipo()
    fig1, ax1 = plt.subplots()
    lh1 = dict()

    fig2, ax2 = plt.subplots()
    lh2 = dict()

    fig3, ax3 = plt.subplots()
    lh3 = dict()

    for x in proto:
        lh1[x[0]] = state1.plot_mean_error_pdr(x[1], ax1, x[2])
        lh2[x[0]] = state1.plot_mean_error_adt(x[1], ax2, x[2])
        lh3[x[0]] = state1.plot_mean_error_jitter(x[1], ax3, x[2])
    for x in sproto:
        lh1[x[0]] = state1.plot_mean_error_pdr_sec(x[1], ax1, x[2])
        lh2[x[0]] = state1.plot_mean_error_adt_sec(x[1], ax2, x[2])
        lh3[x[0]] = state1.plot_mean_error_jitter_sec(x[1], ax3, x[2])

    for ax, lh in zip((ax1, ax2, ax3), (lh1, lh2, lh3)):
        ax.legend((lh.values()), (lh.keys()), loc="best")
        ax.grid()

    ax1.set(xlabel='Vitesse (m/s)', ylabel='Taux de paquets livrés (%)')
    ax2.set(xlabel='Vitesse (m/s)', ylabel='Délai de livraison moyen (ms)')
    ax3.set(xlabel='Vitesse (m/s)', ylabel='Gigue moyenne (ms)')

    for f, m in zip((fig1, fig2, fig3), ("pdr", "adt", "jitter")):
        f.tight_layout()
        f.savefig("topo_" + str(grid) + "_" + m + ".eps", transparent=True)
    plt.close()
