"""
Copyright or © or Copr. University of Nantes / LS2N (2021)

Contributor:
* Alexis Bitaillou

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
"""
import hashlib
import itertools
import json
import multiprocessing
import re
import shutil
import subprocess
from typing import List, Union, Callable

from stats import Stats
from worker_db_handler import lock_test, unlock_test, mark_test_completed
from worker_qualnet_handler import make_qualnet_test, compile_qualnet, change_branch, change_commit


class TestBatcher:
    def __init__(self, config_dir: str, test_name: str, commit_tag: str):
        self.nb_threads = multiprocessing.cpu_count() // 2
        self.config_dir = config_dir  # type: str
        self.test_name = test_name
        self.config_file = self.config_dir + "/" + self.test_name + ".config"  # type: str
        self.app_file = self.config_dir + "/" + self.test_name + ".app"  # type: str
        self.conf_file = self.config_dir + "/qrouting.conf"  # type: str
        self.log_file = self.config_dir + "/log.conf"  # type: str
        self.config = dict()  # type: dict
        self.conf = dict()  # type: dict
        self.app = dict()  # type: dict
        self.last_app_content = ""  # type: str
        self.commit_tag = commit_tag

    def check_free_space(self):
        free_space = shutil.disk_usage("/").free
        if free_space < 500 * 1024 * 1024:
            subprocess.run("rm *.stat", cwd=self.config_dir, stderr=subprocess.STDOUT, shell=True)

    def change_config_value(self, key: str, value):
        pattern = re.compile(key + r' \w+-?\w*')
        with open(self.config_file, 'r', encoding='utf8') as file:
            tmp = file.read()
        result = re.search(pattern, tmp)
        if result is not None:
            tmp = tmp.replace(result.group(0), key + " " + str(value))
            # tmp = tmp.replace("\nOLSRv2-ATTACHED-NETWORK-IPv4 169.0.0.0 24", "")
            print("Writing config file. Don't stop now.")
            with open(self.config_file, 'w', encoding='utf8') as file:
                file.write(tmp)
            print("Writing finished.")
        else:
            print("Pattern not found in config file: ", key, pattern)

    def check_last_app_content(self):
        if self.last_app_content == "":
            with open(self.app_file, 'r',
                      encoding='utf8') as file:
                tmp = file.read()
            self.last_app_content = tmp

    def change_app_bg(self, end_time: str, interval: str):
        pattern = re.compile(
            r"TRAFFIC-GEN \d+ \d+ DET \d+\w{1,2} DET \d+\w{1,2} RND DET \d+ \w{3} \d+\.?\d*\w{1,2} ?\d*\.?\d*\w{0,2} \d\.\d+ ?\d?\.?\d* NOLB")
        with open(self.app_file, 'r',
                  encoding='utf8') as file:
            tmp = file.read()
        result = re.findall(pattern, tmp)
        if len(result) > 0:
            for i in range(len(result)):
                splitted = result[i].split(" ")  # type: List[str]
                if end_time != "":
                    splitted[6] = end_time
                splitted[11] = interval
                unsplit = " ".join(splitted)
                tmp = tmp.replace(result[i], unsplit)
            print("Writing config file. Don't stop now.")
            self.last_app_content = tmp
            with open(self.app_file, 'w', encoding='utf8') as file:
                file.write(tmp)
            print("Writing finished.")
        else:
            print("Traffic-gen pattern not found.", pattern)

    def change_app_fg(self, end_time: str, interval: str):
        pattern = re.compile(r"CBR \d+ \d+ \d+ \d+ \d+\.?\d*\w{1,2} \d+\w{1,2} \d+\w{1,2}")
        with open(self.app_file, 'r',
                  encoding='utf8') as file:
            tmp = file.read()
        result = re.findall(pattern, tmp)
        if len(result) > 0:
            for i in range(len(result)):
                splitted = result[i].split(" ")  # type: List[str]
                if end_time != "":
                    splitted[7] = end_time
                splitted[5] = interval
                unsplit = " ".join(splitted)
                tmp = tmp.replace(result[i], unsplit)
            self.last_app_content = tmp
            print("Writing config file. Don't stop now.")
            with open(self.app_file, 'w', encoding='utf8') as file:
                file.write(tmp)
            print("Writing finished.")
        else:
            print("CBR pattern not found", pattern)

    def change_conf_value(self, key: str, value: Union[float, int, str]):
        pattern = re.compile(key + r"=\d+\.?\d*")
        with open(self.conf_file, 'r', encoding='utf8') as file:
            tmp = file.read()
        result = re.search(pattern, tmp)
        if result is not None:
            tmp = tmp.replace(result.group(0), key + "=" + str(value))
            print("Writing config file. Don't stop now.")
            with open(self.conf_file, 'w', encoding='utf8') as file:
                file.write(tmp)
            print("Writing finished.")
        else:
            print("Adding key %s to qrouting.conf" % key)
            new_tmp = tmp.replace("\n\n", key + "=" + str(value), 1)
            if new_tmp == tmp:
                tmp += "\n%s=%s" % (key, value)
            else:
                tmp = new_tmp
            with open(self.conf_file, 'w', encoding='utf8') as file:
                file.write(tmp)

    def change_log_value(self, key: str, value: str):
        pattern = re.compile(key + r'=[a-zA-Z0-9_\-\.]*')
        with open(self.log_file, 'r',
                  encoding='utf8') as file:
            tmp = file.read()
        result = re.search(pattern, tmp)
        if result is not None:
            tmp = tmp.replace(result.group(0), key + "=" + str(value))
            print("Writing config file. Don't stop now.")
            with open(self.log_file, 'w', encoding='utf8') as file:
                file.write(tmp)
            print("Writing finished.")
        else:
            print("Adding key %s to log.conf" % key)
            new_tmp = tmp.replace("\n\n", key + "=" + str(value), 1)
            if new_tmp == tmp:
                tmp += "\n%s=%s" % (key, value)
            else:
                tmp = new_tmp
            with open(self.log_file, 'w', encoding='utf8') as file:
                file.write(tmp)

    def inner_run(self, bin_dir: str) -> Stats:
        current_setting = dict()
        current_setting["scenario"] = self.test_name
        current_setting["commit"] = self.commit_tag
        qrouting_setting = dict()
        for config_values in itertools.product(*self.config.values()):
            keys = list(self.config.keys())
            for i in range(len(keys)):
                current_setting[keys[i]] = config_values[i]
                self.change_config_value(keys[i], config_values[i])
            for conf_values in itertools.product(*self.conf.values()):
                keys = list(self.conf.keys())
                for i in range(len(keys)):
                    qrouting_setting[keys[i]] = conf_values[i]
                    if current_setting["ROUTING-PROTOCOL-IPv4"] == "QROUTING":
                        self.change_conf_value(keys[i], conf_values[i])
                for cbr_info in itertools.product(list(self.app["CBR"]["INTERVAL"]),
                                                  list(self.app["CBR"]["DURATION"])):
                    current_setting["CBR interval"] = cbr_info[0]
                    current_setting["CBR duration"] = cbr_info[1]
                    if cbr_info[1] == "AUTO":
                        self.change_app_fg(str(int(current_setting["SIMULATION-TIME"].replace("M", "")) - 1) + "M",
                                           cbr_info[0])
                    else:
                        self.change_app_fg(cbr_info[1], cbr_info[0])
                    for traffic_gen_info in itertools.product(list(self.app["TRAFFIC-GEN"]["INTERVAL"]), list(
                            self.app["TRAFFIC-GEN"]["DURATION"])):
                        current_setting["Noise interval"] = traffic_gen_info[0]
                        current_setting["Noise duration"] = traffic_gen_info[1]
                        if traffic_gen_info[1] == "AUTO":
                            self.change_app_bg(
                                str(int(current_setting["SIMULATION-TIME"].replace("M", "")) - 1) + "M",
                                traffic_gen_info[0])
                        else:
                            self.change_app_bg(traffic_gen_info[1], traffic_gen_info[0])
                        hash_set = hashlib.sha1((json.dumps(current_setting, sort_keys=True) + json.dumps(
                            qrouting_setting, sort_keys=True)).encode('utf8')).hexdigest()[0:8]
                        output = '_'.join(map(str, [current_setting["ROUTING-PROTOCOL-IPv4"], self.commit_tag,
                                                    current_setting["SEED"], current_setting["SIMULATION-TIME"],
                                                    current_setting["CBR interval"],
                                                    current_setting["CBR duration"],
                                                    current_setting["Noise interval"],
                                                    current_setting["Noise duration"],
                                                    hash_set]))
                        self.change_log_value("logFileCBR", output + "_cbr")
                        self.change_log_value("logFileRoute", output + "_route")
                        self.change_log_value("logFileDrop", output + "_drop")
                        self.change_log_value("logFileBattery", output + "_battery")
                        self.change_log_value("logFileQTable", output + "_table")
                        self.change_log_value("logFileForwardTableIPv4", output + "_forward")
                        process_time = make_qualnet_test(self.config_dir, self.config_file, output, bin_dir)
                        self.check_last_app_content()
                        yield Stats(self.config_dir + "/" + output + ".stat", current_setting, process_time, hash_set,
                                    qrouting_setting, self.last_app_content)
                if current_setting["ROUTING-PROTOCOL-IPv4"] != "QROUTING":
                    break

    def run(self, bin_dir: str) -> Stats:
        self.check_free_space()
        try:
            yield self.inner_run(bin_dir)
        except FileNotFoundError:
            pass


class SQLTestBatcher(TestBatcher):
    def __init__(self, config_dir: str, source_dir: str, connection: Callable, worker_id: str):
        super().__init__(config_dir, "", "")
        self.base_dir = config_dir
        self.worker_id = worker_id
        self.connection = connection
        self.test = None
        self.source = source_dir

    def run(self, bin_dir: str):
        self.test = lock_test(self.connection(), self.worker_id)
        while self.test is not None:
            self.test_name = self.test[1]  # type: str
            self.config_dir = self.base_dir + "/" + self.test_name
            self.config_file = self.config_dir + "/" + self.test_name + ".config"  # type: str
            self.app_file = self.config_dir + "/" + self.test_name + ".app"  # type: str
            self.conf_file = self.config_dir + "/qrouting.conf"  # type: str
            self.log_file = self.config_dir + "/log.conf"  # type: str
            self.check_free_space()
            self.last_app_content = ""  # type: str
            self.commit_tag = json.loads(self.test[2])[2]  # type: str
            self.config = json.loads(self.test[3])  # type: dict
            for key in self.config.keys():
                self.config[key] = [self.config[key]]
            self.config["ROUTING-PROTOCOL-IPv4"] = [json.loads(self.test[2])[0]]
            self.app = json.loads(self.test[4])
            for key in self.app.keys():
                for inner_key in self.app[key].keys():
                    self.app[key][inner_key] = [self.app[key][inner_key]]
            self.conf = json.loads(self.test[5])
            for key in self.conf.keys():
                self.conf[key] = [self.conf[key]]
            if json.loads(self.test[2])[0] == "QROUTING":
                change_branch(json.loads(self.test[2])[1], self.source)
                change_commit(self.commit_tag, self.source)
                compile_qualnet(self.source)
            try:
                stat = self.inner_run(bin_dir)
                yield stat
            except FileNotFoundError:
                pass
            except Exception as e:
                unlock_test(self.connection(), self.test[0])
                raise e
            mark_test_completed(self.connection(), self.test[0])
            self.test = lock_test(self.connection(), self.worker_id)


class JSONTestBatcher(TestBatcher):
    def __init__(self, config_dir: str, json_conf: str, commit_tag: str, testname: str):
        super().__init__(config_dir, testname, commit_tag)

        def read_json(_filename) -> dict:
            with open(_filename, 'r', encoding="utf-8") as file:
                text = file.read()
                return json.loads(text)

        json_data = read_json(json_conf)
        self.config = json_data["CONFIG"]  # type: dict
        self.conf = json_data["CONF"]  # type: dict
        self.app = json_data["APP"]  # type: dict
        if "PROCESS" in json_data.keys():
            self.nb_threads = json_data["PROCESS"]
